from contextlib import contextmanager
from pathlib import Path
import sqlite3
from typing import Dict, Iterable, Tuple
from appdirs import user_cache_dir
from gersemi.__version__ import __author__, __title__, __version__


def cache_path() -> Path:
    directory = Path(
        user_cache_dir(appname=__title__, appauthor=__author__, version=__version__)
    )
    directory.mkdir(parents=True, exist_ok=True)
    return directory / "cache.db"


def create_tables(cursor: sqlite3.Cursor):
    cursor.execute(
        """
        CREATE TABLE IF NOT EXISTS files (
            path TEXT PRIMARY KEY,
            size INTEGER NOT NULL,
            modification_time INTEGER NOT NULL
        )"""
    )
    cursor.execute(
        """
        CREATE TABLE IF NOT EXISTS formatted (
            path TEXT PRIMARY KEY,
            configuration_summary TEXT NOT NULL,
            FOREIGN KEY (path) REFERENCES files (path)
        )"""
    )


@contextmanager
def database_cursor(path):
    connection = sqlite3.connect(str(path), detect_types=sqlite3.PARSE_DECLTYPES)
    connection.execute("PRAGMA foreign_keys = 1")
    try:
        cursor = connection.cursor()
        create_tables(cursor)
        yield cursor
    finally:
        connection.commit()
        connection.close()


def create_file_entry(path: Path):
    s = path.stat()
    return str(path), s.st_size, s.st_mtime_ns


class Cache:
    def __init__(self, cursor: sqlite3.Cursor):
        self.cursor = cursor

    def store_files(self, configuration_summary: str, files: Iterable[Path]) -> None:
        f = list(files)
        self.cursor.executemany(
            "INSERT OR REPLACE INTO files VALUES (?, ?, ?)",
            map(create_file_entry, f),
        )
        self.cursor.executemany(
            "INSERT OR REPLACE INTO formatted VALUES (?, ?)",
            [(str(path), configuration_summary) for path in f],
        )

    def get_files(self, configuration_summary: str) -> Dict[Path, Tuple[int, int]]:
        return {
            Path(path): (size, modification_time)
            for path, size, modification_time in self.cursor.execute(
                """
                SELECT *
                FROM files
                WHERE files.path IN (
                    SELECT formatted.path
                    FROM formatted
                    WHERE formatted.configuration_summary = (?)
                )""",
                (configuration_summary,),
            )
        }


class DummyCache:
    def store_files(self, *args, **kwargs) -> None:
        pass

    def get_files(self, _) -> Dict[Path, Tuple[int, int, str]]:
        return {}


@contextmanager
def create_cache(enable_cache):
    if not enable_cache:
        yield DummyCache()
        return

    try:
        with database_cursor(cache_path()) as cursor:
            yield Cache(cursor)
    except sqlite3.Error:
        yield DummyCache()



import argparse
from dataclasses import fields
import pathlib
import sys
from lark import __version__ as lark_version
from gersemi.configuration import (
    normalize_definitions,
    normalize_extensions,
    sanitize_list_expansion,
    ControlConfiguration,
    OutcomeConfiguration,
    ListExpansion,
    indent_type,
    workers_type,
)
from gersemi.configuration_reports import default_report
from gersemi.print_config_kind import PrintConfigKind, print_config_kind
from gersemi.return_codes import SUCCESS, FAIL
from gersemi.runner import run, print_to_stderr
from gersemi.__version__ import __title__, __version__

MISSING = "(missing)"


try:
    from colorama import __version__ as colorama_version

except ImportError:
    colorama_version = MISSING


class ShowVersion(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        print(f"{__title__} {__version__}")
        print(f"lark {lark_version}")
        print(f"colorama {colorama_version}")
        print(f"Python {sys.version}")
        sys.exit(SUCCESS)


def toggle_action(predicate):
    class Impl(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            setattr(
                namespace,
                self.dest,
                not predicate(option_string),
            )

    return Impl


toggle_with_no_prefix = toggle_action(lambda s: s.startswith("--no-"))


def create_argparser():
    parser = argparse.ArgumentParser(
        description="A formatter to make your CMake code the real treasure.",
        prog="gersemi",
        add_help=False,
    )
    modes_group = parser.add_argument_group("modes")
    modes_group.add_argument(
        "-c",
        "--check",
        dest="check_formatting",
        action="store_true",
        help=f"""
    Check if files require reformatting.
    Return {SUCCESS} when there's nothing to reformat.
    Return {FAIL} when some files would be reformatted.
    It can be used together with --diff.
            """,
    )
    modes_group.add_argument(
        "-i",
        "--in-place",
        dest="in_place",
        action="store_true",
        help="Format files in-place.",
    )
    modes_group.add_argument(
        "--diff",
        dest="show_diff",
        action="store_true",
        help="""
    Show diff on stdout for each formatted file instead.
    It can be used together with --check.
        """,
    )
    modes_group.add_argument(
        "--print-config",
        dest="print_config",
        choices=[e.value for e in PrintConfigKind],
        help=f"""Print configuration for files.
        {" ".join(map(lambda attr: attr.description, PrintConfigKind))}
        Command line arguments are taken into consideration just
        as they would be for formatting.
        When configuration file is found values in "definitions" are printed as relative
        paths, otherwise absolute paths are printed.
        Output can be placed in .gersemirc file verbatim.""",
    )
    modes_group.add_argument(
        "--version",
        nargs=0,
        dest="show_version",
        action=ShowVersion,
        help="Show version.",
    )
    modes_group.add_argument(
        "-h",
        "--help",
        action="help",
        default=argparse.SUPPRESS,
        help="Show this help message and exit.",
    )

    outcome_conf_doc: dict[str, str] = {
        item.name: item.metadata["description"] for item in fields(OutcomeConfiguration)
    }
    outcome_configuration_group = parser.add_argument_group(
        title="outcome configuration", description=OutcomeConfiguration.__doc__
    )
    outcome_configuration_group.add_argument(
        "-l",
        "--line-length",
        metavar="INTEGER",
        dest="line_length",
        type=int,
        help=f"{outcome_conf_doc['line_length']} [default: {OutcomeConfiguration.line_length}]",
    )
    outcome_configuration_group.add_argument(
        "--indent",
        metavar="(INTEGER | tabs)",
        dest="indent",
        type=indent_type,
        help=f"{outcome_conf_doc['indent']} [default: {repr(OutcomeConfiguration.indent)}]",
    )
    outcome_configuration_group.add_argument(
        "--unsafe",
        dest="unsafe",
        action="store_true",
        default=None,
        help=outcome_conf_doc["unsafe"],
    )
    outcome_configuration_group.add_argument(
        "--definitions",
        dest="definitions",
        metavar="src",
        default=None,
        nargs="+",
        type=pathlib.Path,
        help=outcome_conf_doc["definitions"],
    )
    outcome_configuration_group.add_argument(
        "--list-expansion",
        dest="list_expansion",
        choices=["favour-inlining", "favour-expansion"],
        help=f"""
    {outcome_conf_doc['list_expansion']}
    {" ".join(map(lambda attr: attr.description, ListExpansion))}
    [default: {OutcomeConfiguration.list_expansion.value}]
            """,
    )
    outcome_configuration_group.add_argument(
        "--warn-about-unknown-commands",
        "--no-warn-about-unknown-commands",
        dest="warn_about_unknown_commands",
        action=toggle_with_no_prefix,
        nargs=0,
        default=None,
        help=f"""
    {outcome_conf_doc["warn_about_unknown_commands"]}
    [default: warnings enabled, same as --warn-about-unknown-commands]
        """,
    )
    outcome_configuration_group.add_argument(
        "--disable-formatting",
        "--enable-formatting",
        dest="disable_formatting",
        action=toggle_action(lambda s: s == "--enable-formatting"),
        nargs=0,
        default=None,
        help=f"""
    {outcome_conf_doc["disable_formatting"]}
    [default: formatting enabled]
        """,
    )
    outcome_configuration_group.add_argument(
        "--extensions",
        dest="extensions",
        metavar="extension-name-or-path",
        default=None,
        nargs="+",
        type=str,
        help=outcome_conf_doc["extensions"],
    )

    control_conf_doc: dict[str, str] = {
        item.name: item.metadata["description"] for item in fields(ControlConfiguration)
    }
    control_configuration_group = parser.add_argument_group(
        title="control configuration", description=ControlConfiguration.__doc__
    )
    control_configuration_group.add_argument(
        "-q",
        "--quiet",
        "--no-quiet",
        dest="quiet",
        action=toggle_with_no_prefix,
        nargs=0,
        default=None,
        help=f"{control_conf_doc['quiet']} [default: don't skip, same as --no-quiet]",
    )

    if colorama_version == MISSING:
        warn_about_missing_colorama = " Warning: missing colorama."
    else:
        warn_about_missing_colorama = ""

    control_configuration_group.add_argument(
        "--color",
        "--no-color",
        dest="color",
        action=toggle_with_no_prefix,
        nargs=0,
        default=None,
        help=f"""
        {control_conf_doc['color']}
        {warn_about_missing_colorama}
        [default: don't colorize diff, same as --no-color]
        """,
    )
    control_configuration_group.add_argument(
        "-w",
        "--workers",
        metavar="(INTEGER | max)",
        dest="workers",
        type=workers_type,
        help=f"""
    {control_conf_doc['workers']} [default: {repr(ControlConfiguration.workers)}]
        """,
    )
    control_configuration_group.add_argument(
        "--cache",
        "--no-cache",
        dest="cache",
        action=toggle_with_no_prefix,
        nargs=0,
        default=None,
        help=f"""
    {control_conf_doc["cache"]}
    [default: cache enabled, same as --cache]
        """,
    )
    control_configuration_group.add_argument(
        "--config",
        dest="configuration_file",
        type=pathlib.Path,
        default=None,
        help=f"""
    {control_conf_doc["configuration_file"]}
    [default: omitted]
        """,
    )
    control_configuration_group.add_argument(
        "--warnings-as-errors",
        dest="warnings_as_errors",
        action="store_true",
        default=None,
        help=control_conf_doc["warnings_as_errors"],
    )

    parser.add_argument(
        dest="sources",
        metavar="src",
        nargs="*",
        type=pathlib.Path,
        help="""
    File or directory to format. When directory is provided then CMakeLists.txt
    and files with .cmake extension are automatically discovered.
    If only `-` is provided, input is taken from stdin instead.
            """,
    )

    return parser


def is_stdin_mixed_with_file_input(sources):
    if sources is None:
        return False

    return pathlib.Path("-") in sources and len(sources) != 1


def postprocess_args(args):
    args.sources = set(args.sources)
    if args.definitions is not None:
        args.definitions = set(args.definitions)

    args.definitions = normalize_definitions(args.definitions)
    args.extensions = normalize_extensions(args.extensions)
    args.list_expansion = sanitize_list_expansion(args.list_expansion)
    args.print_config = print_config_kind(args.print_config)


def error(text):
    print_to_stderr(text)
    sys.exit(FAIL)


def main():
    try:
        argparser = create_argparser()
        args = argparser.parse_args()

        if args.print_config == PrintConfigKind.Default.value:
            print(default_report())
            sys.exit(SUCCESS)

        postprocess_args(args)

        if any(map(is_stdin_mixed_with_file_input, [args.sources, args.definitions])):
            error("Don't mix stdin with file input")

        if len(args.sources) < 1:
            sys.exit(SUCCESS)

        sys.exit(run(args))
    except Exception as exception:  # pylint: disable=broad-exception-caught
        error(exception)


if __name__ == "__main__":
    main()



from gersemi.formatted_file import FormattedFile
from gersemi.task_result import TaskResult
from gersemi.tasks.check_formatting import check_formatting
from gersemi.tasks.show_diff import get_diff


def check_and_show_diff(
    should_colorize: bool, formatted_file: FormattedFile
) -> TaskResult:
    result = check_formatting(formatted_file)
    result.to_stdout = get_diff(should_colorize, formatted_file)
    return result



from typing import List
from gersemi.formatted_file import FormattedFile
from gersemi.return_codes import FAIL, SUCCESS
from gersemi.task_result import TaskResult
from gersemi.warnings import FormatterWarning, WrongFormattingWarning


def check_formatting(formatted_file: FormattedFile) -> TaskResult:
    warnings: List[FormatterWarning] = list(formatted_file.warnings)
    if formatted_file.before != formatted_file.after:
        code = FAIL
        warnings.append(WrongFormattingWarning())
    else:
        code = SUCCESS

    return TaskResult(path=formatted_file.path, return_code=code, warnings=warnings)



import argparse
from collections import defaultdict, ChainMap
import collections.abc
from functools import partial
from hashlib import sha1
from itertools import chain
import multiprocessing as mp
import multiprocessing.dummy as mp_dummy
from pathlib import Path
import sys
from typing import Callable, Dict, List, Iterable, Optional, Tuple, Union
from gersemi.cache import create_cache
from gersemi.configuration import (
    Configuration,
    ControlConfiguration,
    find_closest_dot_gersemirc,
    make_control_configuration,
    make_outcome_configuration,
    MaxWorkers,
    max_number_of_workers,
    NotSupportedKeys,
    OutcomeConfiguration,
    Workers,
)
from gersemi.configuration_reports import minimal_report, verbose_report
from gersemi.custom_command_definition_finder import (
    find_custom_command_definitions,
    get_just_definitions,
)
from gersemi.formatted_file import FormattedFile
from gersemi.formatter import create_formatter, NullFormatter, Formatter
from gersemi.mode import get_mode, Mode
from gersemi.parser import PARSER as parser
from gersemi.extensions import load_definitions_from_extensions
from gersemi.print_config_kind import PrintConfigKind
from gersemi.result import Result, Error, apply, get_error_message
from gersemi.return_codes import FAIL, INTERNAL_ERROR, SUCCESS
from gersemi.task_result import TaskResult
from gersemi.tasks.check_formatting import check_formatting
from gersemi.tasks.check_and_show_diff import check_and_show_diff
from gersemi.tasks.do_nothing import do_nothing
from gersemi.tasks.forward_to_stdout import forward_to_stdout
from gersemi.tasks.format_file import format_file
from gersemi.tasks.rewrite_in_place import rewrite_in_place
from gersemi.tasks.show_diff import show_diff
from gersemi.utils import fromfile, smart_open
from gersemi.keywords import Keywords
from gersemi.warnings import UnknownCommandWarning


CHUNKSIZE = 16


print_to_stdout = partial(print, file=sys.stdout, end="")
print_to_stderr = partial(print, file=sys.stderr)


class WarningSink:
    def __init__(self, quiet):
        self.quiet = quiet
        self.at_least_one_warning_issued = False
        self.records = []

    def __call__(self, s: str):
        self.at_least_one_warning_issued = True
        if not self.quiet:
            self.records.append(s)

    def flush(self):
        for s in self.records:
            print_to_stderr(s)


class StatusCode:
    def __init__(self):
        self.value = SUCCESS

    def __iadd__(self, other):
        if isinstance(other, int):
            self.value = max(self.value, other)
            return self

        if isinstance(other, collections.abc.Iterable):
            for item in other:
                self += item

            return self

        raise RuntimeError(f"Invalid type: {type(other)}")


def get_files(paths: Iterable[Path]) -> Iterable[Path]:
    def get_files_from_single_path(path):
        if path.is_dir():
            return chain(path.rglob("CMakeLists.txt"), path.rglob("*.cmake"))
        return [path]

    return sorted(
        set(
            item.resolve(True) if item != Path("-") else item
            for path in paths
            for item in get_files_from_single_path(path)
        )
    )


def has_custom_command_definition(code: str) -> bool:
    lowercased = code.lower()
    has_function_definition = "function" in lowercased and "endfunction" in lowercased
    has_macro_definition = "macro" in lowercased and "endmacro" in lowercased
    return has_function_definition or has_macro_definition


def find_custom_command_definitions_in_file_impl(filepath: Path) -> Dict[str, Keywords]:
    with smart_open(filepath, "r") as f:
        code = f.read()
    if not has_custom_command_definition(code):
        return {}

    parse_tree = parser.parse(code)
    return find_custom_command_definitions(parse_tree, filepath)


def find_custom_command_definitions_in_file(
    filepath: Path,
) -> Result[Dict[str, Keywords]]:
    return apply(find_custom_command_definitions_in_file_impl, filepath)


def check_conflicting_definitions(definitions, warning_sink: WarningSink):
    for name, info in definitions.items():
        if len(info) > 1:
            warning_sink(f"Warning: conflicting definitions for '{name}':")
            places = sorted(where for _, where in info)
            for index, where in enumerate(places):
                kind = "(used)   " if index == 0 else "(ignored)"
                warning_sink(f"{kind} {where}")


def find_all_custom_command_definitions(
    paths: Iterable[Path], pool, warning_sink: WarningSink
) -> Dict[str, Keywords]:
    result: Dict = {}

    try:
        files = get_files(paths)
    except FileNotFoundError as e:
        # pylint: disable=broad-exception-raised
        raise Exception(f"Definition path doesn't exist: {e.filename}") from e

    find = find_custom_command_definitions_in_file

    for defs in pool.imap_unordered(find, files, chunksize=CHUNKSIZE):
        if isinstance(defs, Error):
            warning_sink(get_error_message(defs))
            continue

        for name, info in defs.items():
            if name in result:
                result[name].extend(info)
            else:
                result[name] = info

    check_conflicting_definitions(result, warning_sink)

    return get_just_definitions(result)


def select_task(mode: Mode, configuration: Configuration):
    return {
        Mode.ForwardToStdout: forward_to_stdout,
        Mode.RewriteInPlace: rewrite_in_place,
        Mode.CheckFormatting: check_formatting,
        Mode.ShowDiff: partial(show_diff, configuration.control.color),
        Mode.CheckFormattingAndShowDiff: partial(
            check_and_show_diff, configuration.control.color
        ),
    }[mode]


def run_task(
    path: Path,
    formatter: Union[NullFormatter, Formatter],
    task: Callable[[FormattedFile], TaskResult],
) -> TaskResult:
    formatted_file: Result[FormattedFile] = apply(format_file, path, formatter)
    if isinstance(formatted_file, Error):
        return TaskResult(
            path=path,
            return_code=INTERNAL_ERROR,
            to_stderr=get_error_message(formatted_file),
        )
    return task(formatted_file)


def consume_task_result(
    task_result: TaskResult,
    configuration: Configuration,
    warning_sink: WarningSink,
) -> Tuple[Path, int, bool]:
    if task_result.to_stdout != "":
        print_to_stdout(task_result.to_stdout)

    warnings = (
        [w for w in task_result.warnings if not isinstance(w, UnknownCommandWarning)]
        if not configuration.outcome.warn_about_unknown_commands
        else task_result.warnings
    )

    for warning in warnings:
        warning_sink(warning.get_message(fromfile(task_result.path)))

    if task_result.to_stderr != "":
        warning_sink(task_result.to_stderr)

    return task_result.path, task_result.return_code, (len(warnings) > 0)


def create_pool(is_stdin_in_sources, workers: Workers):
    if is_stdin_in_sources:
        return mp_dummy.Pool

    if isinstance(workers, MaxWorkers):
        value = max_number_of_workers()
    else:
        if workers <= 1:
            return mp_dummy.Pool

        value = workers

    return partial(mp.Pool, processes=value)


def summarize_configuration(configuration, extension_definitions):
    hasher = sha1()
    hasher.update(repr(configuration).encode("utf-8"))
    hasher.update(repr(extension_definitions).encode("utf-8"))
    return hasher.hexdigest()


def split_files_by_formatting_state(
    cache,
    configuration: OutcomeConfiguration,
    files: Iterable[Path],
):
    extension_definitions = load_definitions_from_extensions(configuration.extensions)
    summary = summarize_configuration(configuration, extension_definitions)
    known_files = cache.get_files(summary)
    already_formatted_files = []
    files_to_format = []
    for f in files:
        if f not in known_files:
            files_to_format.append(f)
        else:
            s = f.stat()
            if (s.st_size, s.st_mtime_ns) != known_files[f]:
                files_to_format.append(f)
            else:
                already_formatted_files.append(f)

    return already_formatted_files, files_to_format


def store_files_in_cache(
    mode: Mode, cache, configuration_summary: str, files: Iterable[Path]
) -> None:
    if mode in [
        Mode.CheckFormatting,
        Mode.CheckFormattingAndShowDiff,
        Mode.RewriteInPlace,
    ]:
        cache.store_files(configuration_summary, files)


def select_task_for_already_formatted_files(mode: Mode):
    return {
        Mode.ForwardToStdout: forward_to_stdout,
    }.get(mode, do_nothing)


def handle_already_formatted_files(
    mode: Mode,
    configuration: Configuration,
    warning_sink: WarningSink,
    already_formatted_files: Iterable[Path],
) -> Iterable[int]:
    task = select_task_for_already_formatted_files(mode)
    formatter = NullFormatter()
    execute = partial(run_task, formatter=formatter, task=task)
    results = [
        consume_task_result(result, configuration, warning_sink)
        for result in map(execute, already_formatted_files)
    ]
    return (code for _, code, _ in results)


def handle_files_to_format(  # pylint: disable=too-many-arguments,too-many-positional-arguments
    mode: Mode,
    configuration: Configuration,
    cache,
    pool,
    warning_sink: WarningSink,
    files_to_format: Iterable[Path],
) -> Iterable[int]:
    custom_command_definitions = find_all_custom_command_definitions(
        set(configuration.outcome.definitions), pool, warning_sink
    )
    extension_definitions = load_definitions_from_extensions(
        configuration.outcome.extensions
    )

    formatter = create_formatter(
        configuration.outcome,
        ChainMap(custom_command_definitions, extension_definitions),
    )
    task = select_task(mode, configuration)
    execute = partial(run_task, formatter=formatter, task=task)

    results = [
        consume_task_result(result, configuration, warning_sink)
        for result in pool.imap_unordered(execute, files_to_format, chunksize=CHUNKSIZE)
    ]
    store_files_in_cache(
        mode,
        cache,
        summarize_configuration(configuration.outcome, extension_definitions),
        (
            path
            for path, code, has_warnings in results
            if code == SUCCESS and path != Path("-") and (not has_warnings)
        ),
    )
    return (code for _, code, _ in results)


class GetConfiguration:
    def __init__(
        self,
        args: argparse.Namespace,
        control: ControlConfiguration,
        warning_sink: WarningSink,
    ):
        self.args = args
        self.control = control
        self.processed_configuration_files: List[Path] = []
        self.warning_sink = warning_sink

    def _warn(self, item: NotSupportedKeys, text: str):
        self.warning_sink(f"{item.path}: {text}")

    def _inform_about_not_supported_keys(self, item: NotSupportedKeys):
        if item.path in self.processed_configuration_files:
            return

        if item.path is None:
            return

        self.processed_configuration_files.append(item.path)

        if len(item.command_line_only) > 0:
            keys = ", ".join(sorted(item.command_line_only))
            self._warn(
                item, f"these options are supported only through command line: {keys}"
            )

        if len(item.unknown) > 0:
            keys = ", ".join(sorted(item.unknown))
            self._warn(item, f"these options are not supported: {keys}")

    def __call__(self, configuration_file: Optional[Path]) -> Configuration:
        outcome_configuration, not_supported_keys = make_outcome_configuration(
            configuration_file=configuration_file,
            args=self.args,
        )
        self._inform_about_not_supported_keys(not_supported_keys)
        return Configuration(control=self.control, outcome=outcome_configuration)


def split_files_by_configuration_file(
    paths: Iterable[Path], control: ControlConfiguration
):
    if control.configuration_file is not None:
        return {control.configuration_file: paths}

    result = defaultdict(list)
    for path in paths:
        result[find_closest_dot_gersemirc(path)].append(path)

    return result


def print_configuration_report(
    kind: PrintConfigKind,
    buckets: Dict[Optional[Path], Iterable[Path]],
    get_configuration: GetConfiguration,
):
    report = {
        PrintConfigKind.Minimal: lambda conf_file, conf, _: minimal_report(
            conf_file, conf
        ),
        PrintConfigKind.Verbose: verbose_report,
    }.get(kind, lambda *args: None)

    for config_file, target_files in buckets.items():
        config = get_configuration(config_file)
        result = report(config_file, config.outcome, target_files)
        if result is not None:
            print_to_stdout(result)


# pylint: disable=too-many-locals
def run(args: argparse.Namespace):
    try:
        requested_files = get_files(args.sources)
    except FileNotFoundError as e:
        # pylint: disable=broad-exception-raised
        raise Exception(f"Source path doesn't exist: {e.filename}") from e

    mode = get_mode(args)
    control = make_control_configuration(args)
    warning_sink = WarningSink(control.quiet)
    pool_cm = create_pool(
        is_stdin_in_sources=(Path("-") in requested_files),
        workers=control.workers,
    )

    buckets = split_files_by_configuration_file(requested_files, control)
    get_configuration = GetConfiguration(args, control, warning_sink)
    if mode == Mode.PrintConfig:
        print_configuration_report(args.print_config, buckets, get_configuration)
        return SUCCESS

    with create_cache(control.cache) as cache, pool_cm() as pool:
        status_code = StatusCode()
        for config_file, files in buckets.items():
            config = get_configuration(config_file)
            if config.outcome.disable_formatting:
                continue

            already_formatted_files, files_to_format = split_files_by_formatting_state(
                cache, config.outcome, files
            )

            status_code += handle_already_formatted_files(
                mode, config, warning_sink, already_formatted_files
            )
            status_code += handle_files_to_format(
                mode, config, cache, pool, warning_sink, files_to_format
            )

        status_code += (
            FAIL
            if (control.warnings_as_errors and warning_sink.at_least_one_warning_issued)
            else SUCCESS
        )
        warning_sink.flush()
        return status_code.value



from lark import Discard, Tree
from lark.visitors import Transformer
from gersemi.exceptions import ASTMismatch


class DropWhitespaces(Transformer):
    def _drop_node(self, _):
        return Discard

    def line_comment(self, children):
        if not children:
            return Tree("line_comment", [])

        comment_content, *_ = children
        return Tree("line_comment", [comment_content.rstrip()])

    def command_invocation(self, children):
        command_name, *rest = children
        return Tree("command_invocation", [command_name.lower(), *rest])

    def non_command_element(self, children):
        if len(children) == 0:
            return Discard
        return Tree("non_command_element", children)

    def arguments(self, children):
        return Tree("arguments", set(children))

    NEWLINE = _drop_node
    newline_or_gap = _drop_node


def drop_whitespaces(tree):
    return DropWhitespaces(visit_tokens=True).transform(tree)


def check_abstract_syntax_trees_equivalence(lhs, rhs):
    preprocess = drop_whitespaces
    if preprocess(lhs) != preprocess(rhs):
        raise ASTMismatch


def check_code_equivalence(parser, lhs_parsed, rhs):
    rhs_parsed = parser.parse(rhs)
    check_abstract_syntax_trees_equivalence(lhs_parsed, rhs_parsed)



# pylint: disable=too-many-lines
from .argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)


class SetProperty(ArgumentAwareCommandInvocationDumper):
    options = ["GLOBAL", "APPEND", "APPEND_STRING"]
    multi_value_keywords = [
        "TARGET",
        "SOURCE",
        "INSTALL",
        "TEST",
        "CACHE",
        "PROPERTY",
        "TARGET_DIRECTORIES",
        "DIRECTORY",
    ]
    _keyword_formatters = {"PROPERTY": "_format_property"}

    def _format_property(self, args):
        result = self._try_to_format_into_single_line(args, prefix="", postfix="")
        if result is not None:
            return result

        name, *rest = args
        with self.indented():
            formatted_rest = "\n".join(map(self.visit, rest))
        return f"{self.visit(name)}\n{formatted_rest}"


set_property = {"set_property": SetProperty}



from collections import ChainMap
from itertools import dropwhile
from typing import List
from lark import Discard, Tree, Token
from lark.visitors import Transformer_InPlace
from gersemi.ast_helpers import (
    is_commented_argument,
    is_newline,
    is_quoted_argument,
    is_unquoted_argument,
)
from gersemi.builtin_commands import _builtin_commands
from gersemi.types import Nodes


class RemoveSuperfluousEmptyLines(Transformer_InPlace):
    def _drop_edge_empty_lines(self, children) -> List:
        while len(children) > 0 and is_newline(children[-1]):
            children.pop()
        return list(dropwhile(is_newline, children))

    def file(self, children) -> Tree:
        return Tree("file", self._drop_edge_empty_lines(children))

    def block_body(self, children) -> Tree:
        return Tree("block_body", self._drop_edge_empty_lines(children))


class SimplifyParseTree(Transformer_InPlace):
    def non_command_element(self, children: Nodes):
        if len(children) == 0:
            return Discard
        return Tree("non_command_element", children)


class PreserveCustomCommandFormatting(Transformer_InPlace):
    def __init__(self, code, known_definitions):
        super().__init__()
        self.code = code
        self.known_definitions = (
            _builtin_commands
            if known_definitions is None
            else ChainMap(known_definitions, _builtin_commands)
        )

    def _get_original_formatting(self, lparen, rparen):
        start, end = lparen.end_pos, rparen.start_pos
        content = self.code[start:end]
        return Tree("formatted_node", [content])

    def _get_indentation(self, identifier):
        end = identifier.start_pos
        start = end - identifier.column + 1
        return self.code[start:end]

    def _make_custom_command(self, children):
        identifier, lparen, arguments, rparen = children
        indentation = self._get_indentation(identifier)
        return Tree(
            "custom_command",
            [
                indentation,
                identifier,
                arguments,
                self._get_original_formatting(lparen, rparen),
            ],
        )

    def command_invocation(self, children):
        identifier, _, arguments, _ = children
        if identifier.lower() in self.known_definitions:
            return Tree("command_invocation", [identifier, arguments])

        return super().transform(self._make_custom_command(children))


class RecognizeUnquotedLegacyArgument(Transformer_InPlace):
    def _valid_pair(self, lhs, rhs):
        if is_commented_argument(rhs):
            argument, *__ = rhs.children
            return self._valid_pair(lhs, argument)

        if not is_unquoted_argument(lhs):
            return False

        if not (is_unquoted_argument(rhs) or is_quoted_argument(rhs)):
            return False

        end = lhs.children[-1].end_pos
        start = rhs.children[0].start_pos
        return end == start

    def _join_tokens(self, lhs, rhs):
        return Token(
            lhs.children[0].type,
            "".join(lhs.children) + "".join(rhs.children),
            start_pos=lhs.children[0].start_pos,
            end_pos=rhs.children[-1].end_pos,
        )

    def _join_pair(self, lhs, rhs):
        if is_commented_argument(rhs):
            argument, *rest = rhs.children
            return Tree("commented_argument", [self._join_pair(lhs, argument), *rest])

        return Tree("unquoted_argument", [self._join_tokens(lhs, rhs)])

    def _join_arguments(self, children):
        if len(children) < 2:
            return children

        new_children = []
        previous = children[0]
        for current in children[1:]:
            if not self._valid_pair(previous, current):
                new_children.append(previous)
                previous = current
                continue

            previous = self._join_pair(previous, current)

        new_children.append(previous)
        return new_children

    def arguments(self, children):
        original = children
        new_children = self._join_arguments(children)
        while original != new_children:
            original, new_children = new_children, self._join_arguments(new_children)

        return Tree("arguments", new_children)


class PostProcessorStageOne(
    PreserveCustomCommandFormatting,
    SimplifyParseTree,
    RemoveSuperfluousEmptyLines,
    RecognizeUnquotedLegacyArgument,
):
    pass


class SimplifyQuotedArguments(Transformer_InPlace):
    def quoted_argument(self, children):
        return Tree("quoted_argument", ["".join(children[1:-1])])


def postprocess(code, definitions, tree):
    stage_one = PostProcessorStageOne(code, definitions)
    stage_two = SimplifyQuotedArguments()
    return stage_two.transform(stage_one.transform(tree))



from collections import abc


class ImmutableDict(abc.Mapping):
    def __init__(self, d):
        self._d = d

    def __getitem__(self, key):
        return self._d[key]

    def __iter__(self):
        return iter(self._d)

    def __len__(self):
        return len(self._d)

    def __contains__(self, item):
        return item in self._d

    def __hash__(self):
        return hash(frozenset(self._d.items()))


def make_immutable(thing):
    if isinstance(thing, abc.Mapping):
        return ImmutableDict(
            {key: make_immutable(value) for key, value in thing.items()}
        )

    if isinstance(thing, str):
        return thing

    if isinstance(thing, abc.Collection):
        return tuple(map(make_immutable, thing))

    return thing



from enum import Enum
from typing import Optional, Union


class KeywordFormatter(Enum):
    CommandLine = "command_line"
    Pairs = "pairs"


class KeywordPreprocessor(Enum):
    Sort = "sort"
    Unique = "unique"
    SortAndUnique = "sort+unique"


KeywordKind = Union[KeywordFormatter, KeywordPreprocessor]


def get_formatter_kind(
    kind: Union[None, str, KeywordFormatter],
) -> Optional[KeywordFormatter]:
    if isinstance(kind, KeywordFormatter):
        return kind

    return (
        KeywordFormatter(kind) if kind in [e.value for e in KeywordFormatter] else None
    )


def kind_to_formatter(kind: Union[None, str, KeywordFormatter]) -> Optional[str]:
    proper_kind = get_formatter_kind(kind)
    if proper_kind is None:
        return None

    return {
        KeywordFormatter.CommandLine: "_format_command_line",
        KeywordFormatter.Pairs: "_format_keyword_with_pairs",
    }.get(proper_kind, None)


def get_preprocessor_kind(
    kind: Union[None, str, KeywordPreprocessor],
) -> Optional[KeywordPreprocessor]:
    if isinstance(kind, KeywordPreprocessor):
        return kind

    return (
        KeywordPreprocessor(kind)
        if kind in [e.value for e in KeywordPreprocessor]
        else None
    )


def kind_to_preprocessor(kind: Union[None, str, KeywordPreprocessor]) -> Optional[str]:
    proper_kind = get_preprocessor_kind(kind)
    if proper_kind is None:
        return None

    return {
        KeywordPreprocessor.Sort: "_sort_arguments",
        KeywordPreprocessor.Unique: "_keep_unique_arguments",
        KeywordPreprocessor.SortAndUnique: "_sort_and_keep_unique_arguments",
    }.get(proper_kind, None)



# pylint: disable=redefined-outer-name
from functools import partial
import pytest
from tests.fixtures.app import success


@pytest.fixture(scope="function")
def app(app, testfiles):
    extension = (testfiles / "overrides" / "extension.py").resolve()
    return partial(app, "--extensions", extension)


@pytest.fixture(scope="function")
def bad_file(testfiles):
    return (testfiles / "overrides" / "bad.cmake").resolve()


@pytest.fixture(scope="function")
def good_file(testfiles):
    return (testfiles / "overrides" / "good.cmake").resolve()


def test_extension_can_override_builtins(app, bad_file, good_file):
    assert app("-i", "--", bad_file) == success(stdout="", stderr="")

    with open(bad_file, "r", encoding="utf-8") as bad_f, open(
        good_file, "r", encoding="utf-8"
    ) as good_f:
        bad = bad_f.read()
        good = good_f.read()

        assert bad == good



from gersemi.enum_with_metadata import EnumWithMetadata


class PrintConfigKind(EnumWithMetadata):
    Minimal = dict(
        value="minimal",
        description="""
    With "minimal" prints source of outcome configuration (configuration file or defaults)
    and the options that differ from defaults.
        """,
        title="Minimal",
    )
    Verbose = dict(
        value="verbose",
        description="""
    With "verbose" prints source of outcome configuration (configuration file or defaults),
    files for which this configuration is applicable and complete listing of options.
        """,
        title="Verbose",
    )
    Default = dict(
        value="default",
        description="""
    With "default" prints outcome configuration with default values.
        """,
        title="Default",
    )


def print_config_kind(s) -> PrintConfigKind:
    if s in [e.value for e in PrintConfigKind]:
        return PrintConfigKind(s)

    return s



import pytest
from gersemi.exceptions import ParsingError
from .tests_generator import generate_input_only_tests


def test_profile_parser_with_postprocessing(parser_with_postprocessing, case):
    try:
        parser_with_postprocessing.parse(case.content)
    except ParsingError:
        pytest.fail("invalid input to parse")
        raise


pytest_generate_tests = generate_input_only_tests(
    where="formatter", input_extension=".in.cmake"
)



from copy import deepcopy
from typing import Tuple
from gersemi.configuration import OutcomeConfiguration
from gersemi.dumper import Dumper
from gersemi.noop import noop
from gersemi.parser import BARE_PARSER
from gersemi.postprocessor import postprocess
from gersemi.sanity_checker import check_code_equivalence
from gersemi.warnings import FormatterWarnings


class Formatter:
    def __init__(
        self, configuration: OutcomeConfiguration, sanity_checker, known_definitions
    ):
        self.configuration = configuration
        self.sanity_checker = sanity_checker
        self.known_definitions = known_definitions

    def format(self, code) -> Tuple[str, FormatterWarnings]:
        tree = BARE_PARSER.parse(code)
        original = deepcopy(tree)
        dumper = Dumper(self.configuration, self.known_definitions)
        result = dumper.visit(postprocess(code, self.known_definitions, tree))
        self.sanity_checker(BARE_PARSER, original, result)
        return result, dumper.get_warnings()


class NullFormatter:
    def format(self, code):
        return code, []


def create_formatter(configuration: OutcomeConfiguration, known_definitions):
    sanity_checker = check_code_equivalence if not configuration.unsafe else noop
    return Formatter(
        configuration,
        sanity_checker,
        known_definitions,
    )



from gersemi.command_line_formatter import CommandLineFormatter
from gersemi.keyword_kind import KeywordFormatter
from .argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)


class AddCustomTarget(CommandLineFormatter, ArgumentAwareCommandInvocationDumper):
    front_positional_arguments = ["Name"]
    options = ["ALL", "VERBATIM", "USES_TERMINAL", "COMMAND_EXPAND_LISTS"]
    one_value_keywords = [
        "WORKING_DIRECTORY",
        "COMMENT",
        "JOB_POOL",
        "JOB_SERVER_AWARE",
    ]
    multi_value_keywords = ["COMMAND", "DEPENDS", "BYPRODUCTS", "SOURCES"]
    keyword_formatters = {"COMMAND": KeywordFormatter.CommandLine}

    def positional_arguments(self, tree):
        if len(tree.children) > 1:
            return super()._format_command_line(tree.children)
        return super().positional_arguments(tree)


add_custom_target = {"add_custom_target": AddCustomTarget}



# pylint: disable=attribute-defined-outside-init,redefined-outer-name
from functools import partial
import json
import pytest
from tests.fixtures.app import success, fail
from tests.test_executable_print_config import ignore_schema


def extension(implementations):
    return {
        "implementation_present": True,
        "implementation_passes_verification": True,
        "implementations": implementations,
    }


BAD_EXTENSION = {"implementation_present": True}
GOOD_EXTENSION = extension(["add_constellation", "add_nebula"])


@pytest.fixture(scope="function")
def extensions_directory():
    return "extension_exists"


@pytest.fixture(scope="function")
def fake_extension():
    return {}


@pytest.fixture(scope="function")
def app(extensions_directory, fake_extension, testfiles, app):
    return lambda *args, **kwargs: app(
        *args,
        "--",
        testfiles / "extensions" / "there_are_spoilers_ahead",
        env={
            "PYTHONPATH": str(testfiles / "extensions" / extensions_directory),
            "GERSEMI_FAKE_EXTENSION_SETTINGS": json.dumps(fake_extension),
        },
        **kwargs,
    )


@pytest.mark.parametrize("extensions_directory", ["no_extensions_directory"])
def test_extensions_directory_dont_exist(app):
    assert app("--check", "--extensions", "foo") == fail(
        stderr="Missing extension foo\n"
    )

    assert app("--check", "--extensions", "baz", "foo", "bar") == fail(
        stderr="""Missing extension bar
Missing extension baz
Missing extension foo
"""
    )


@pytest.fixture(scope="function")
def kim_dokja_company_as_module():
    return {
        "name": "kim_dokja_company",
        "qualified_name": "gersemi_kim_dokja_company",
    }


@pytest.fixture(scope="function")
def kim_dokja_company_as_file(testfiles):
    value = str(testfiles / "extensions" / "kim_dokja_company_as_file.py")
    return {"name": value, "qualified_name": value}


@pytest.mark.parametrize(
    "extension_fixture",
    [
        "kim_dokja_company_as_module",
        "kim_dokja_company_as_file",
    ],
)
class TestExtensionExists:
    @pytest.fixture(autouse=True)
    def _fixtures(self, app, testfiles, request, extension_fixture):
        extension = request.getfixturevalue(extension_fixture)
        self.extension_name = extension["name"]
        self.extension_qualified_name = extension["qualified_name"]
        self.app = partial(app, "--check", "--extensions", self.extension_name)

        base_directory = (
            testfiles
            / "extensions"
            / "there_are_spoilers_ahead"
            / "you_have_been_warned"
            / "one_last_chance"
        )
        self.correct_formatting = base_directory / "correct_formatting.cmake"
        self.wrong_formatting = base_directory / "wrong_formatting.cmake"

    def test_doesnt_have_command_definitions(self):
        assert self.app() == fail(
            stderr=f"Extension {self.extension_name} doesn't implement command_definitions\n"
        )

    @pytest.mark.parametrize("fake_extension", [BAD_EXTENSION])
    def test_fails_verification(self):
        assert self.app() == fail(
            stderr=f"""Verification failed for extension {self.extension_name}:
{self.extension_qualified_name}:command_definitions: is not a mapping
"""
        )

    @pytest.mark.parametrize("fake_extension", [extension([])])
    def test_warn_about_unknown_commands(self):
        assert self.app() == success(
            stderr=f"""Warning: unknown command 'add_nebula' used at:
{self.correct_formatting}:1:1

Warning: unknown command 'add_constellation' used at:
{self.correct_formatting}:3:1
{self.correct_formatting}:12:1
{self.correct_formatting}:14:1

Warning: unknown command 'add_nebula' used at:
{self.wrong_formatting}:1:1

Warning: unknown command 'add_constellation' used at:
{self.wrong_formatting}:6:1
{self.wrong_formatting}:16:1
{self.wrong_formatting}:23:1

"""
        )

    @pytest.mark.parametrize("fake_extension", [extension(["add_constellation"])])
    def test_only_some_commands_are_implemented(self):
        assert self.app() == fail(
            stderr=f"""Warning: unknown command 'add_nebula' used at:
{self.correct_formatting}:1:1

Warning: unknown command 'add_nebula' used at:
{self.wrong_formatting}:1:1

{self.wrong_formatting} would be reformatted
""",
        )

    @pytest.mark.parametrize("fake_extension", [GOOD_EXTENSION])
    def test_all_commands_are_implemented(self):
        assert self.app() == fail(
            stderr=f"""{self.wrong_formatting} would be reformatted
""",
        )


@pytest.mark.parametrize(
    ["fake_extension", "outcome"],
    [
        (
            GOOD_EXTENSION,
            """## - [kim_dokja_company] additional recognized commands:
##   - add_constellation
##   - add_nebula
##""",
        ),
        (
            BAD_EXTENSION,
            """## - [kim_dokja_company] error:
##   Verification failed for extension kim_dokja_company:
##   gersemi_kim_dokja_company:command_definitions: is not a mapping
##""",
        ),
    ],
)
def test_print_config_verbose_with_extensions(app, testfiles, outcome):
    base = testfiles / "extensions" / "there_are_spoilers_ahead"
    applicable_files = [
        base / "CMakeLists.txt",
        base / "you_have_been_warned" / "CMakeLists.txt",
        base / "you_have_been_warned" / "one_last_chance" / "correct_formatting.cmake",
        base / "you_have_been_warned" / "one_last_chance" / "wrong_formatting.cmake",
    ]
    file_listing = "".join(f"\n## - {f.resolve()}" for f in applicable_files)

    expected_stdout = f"""## Outcome configuration based on defaults,
## it's applicable to these files:{file_listing}
##
## About extensions:
## - [yoo_joonghyuk_company] error:
##   Missing extension yoo_joonghyuk_company
##
{outcome}
## - [han_sooyoung_company] error:
##   Missing extension han_sooyoung_company
##

definitions: []
disable_formatting: false
extensions:
- yoo_joonghyuk_company
- kim_dokja_company
- han_sooyoung_company
indent: 4
line_length: 80
list_expansion: favour-inlining
unsafe: false
warn_about_unknown_commands: true
"""

    assert app(
        "--extensions",
        "yoo_joonghyuk_company",
        "kim_dokja_company",
        "han_sooyoung_company",
        "--print-config",
        "verbose",
    ) == success(stdout=ignore_schema(expected_stdout))



# mypy: disable-error-code="no-redef"
from dataclasses import asdict
import json
import os
import pytest
from gersemi.configuration import (
    OutcomeConfiguration,
    ListExpansion,
    make_configuration_file,
    Tabs,
)

try:
    from pydantic.json_schema import GenerateJsonSchema
    from pydantic.version import version_short
    from pydantic import TypeAdapter
except ImportError:

    class GenerateJsonSchema:
        pass

    class TypeAdapter:
        pass

    def version_short() -> str:
        return "0.0"


def pydantic_version_as_tuple():
    return tuple(map(int, version_short().split(".")))


def get_representation(represented_type):
    return {
        "oneOf": [
            {
                "description": item.description,
                "enum": [item.value],
                "title": item.title,
            }
            for item in represented_type
        ],
        "type": "string",
    }


class CustomizedGenerateJsonSchema(GenerateJsonSchema):
    def generate(self, schema, mode=None):
        result = super().generate(schema, mode=mode)
        result["$schema"] = "https://json-schema.org/draft-07/schema"
        result["$defs"]["ListExpansion"] = get_representation(ListExpansion)
        result["$defs"]["Tabs"] = get_representation(Tabs)

        result["properties"]["indent"]["anyOf"][0]["minimum"] = 1

        return result


@pytest.mark.skipif(
    pydantic_version_as_tuple() < (2, 9), reason="At least pydantic 2.9 is required"
)
def test_schema_in_repository_is_consistent_with_configuration_definition():
    this_file_dir = os.path.dirname(os.path.realpath(__file__))
    schema_path = os.path.join(this_file_dir, "configuration.schema.json")
    with open(schema_path, "r", encoding="utf-8") as f:
        schema = json.load(f)

    assert schema == TypeAdapter(OutcomeConfiguration).json_schema(
        schema_generator=CustomizedGenerateJsonSchema,
    )


def test_example_file_in_repository_is_consistent_with_configuration_definition():
    this_file_dir = os.path.dirname(os.path.realpath(__file__))
    example_path = os.path.join(this_file_dir, ".gersemirc.example")
    with open(example_path, "r", encoding="utf-8") as f:
        example = f.read()

    configuration = OutcomeConfiguration()
    assert example == make_configuration_file(
        asdict(configuration), add_schema_link=True
    )



from gersemi.formatted_file import FormattedFile
from gersemi.return_codes import SUCCESS
from gersemi.task_result import TaskResult


def do_nothing(formatted_file: FormattedFile) -> TaskResult:
    return TaskResult(
        path=formatted_file.path,
        return_code=SUCCESS,
        to_stdout="",
    )



from typing import Sequence, Tuple, Union
from lark import Tree
from lark.visitors import Transformer_InPlace
from gersemi.ast_helpers import is_keyword, is_comment
from gersemi.base_command_invocation_dumper import BaseCommandInvocationDumper
from gersemi.keywords import AnyMatcher


class IsolateTwoWordKeywords(Transformer_InPlace):
    def __init__(self, lhs, rhs):
        super().__init__()
        self.lhs = lhs
        self.rhs = rhs

    def _is_lhs(self, node):
        return is_keyword(self.lhs, node)

    def _is_rhs(self, node):
        return is_keyword(self.rhs, node)

    def arguments(self, children):
        if len(children) <= 1:
            return Tree("arguments", children)

        new_children = []
        accumulator = []
        iterator = iter(children)
        for child in iterator:
            if len(accumulator) > 0:
                if is_comment(child):
                    accumulator.append(child)
                elif self._is_rhs(child):
                    new_children.append(Tree("keyword_argument", [*accumulator, child]))
                    accumulator = []
                else:
                    new_children.extend(accumulator)
                    accumulator = [child]
            else:
                if self._is_lhs(child):
                    accumulator = [child]
                else:
                    new_children.append(child)

        if len(accumulator) > 0:
            new_children.extend(accumulator)

        return Tree("arguments", new_children)


class TwoWordKeywordIsolator(BaseCommandInvocationDumper):
    _two_words_keywords: Sequence[Tuple[str, Union[str, AnyMatcher]]] = []

    def _preprocess_arguments(self, arguments):
        preprocessed = arguments
        for lhs, rhs in self._two_words_keywords:
            preprocessed = IsolateTwoWordKeywords(lhs, rhs).transform(preprocessed)
        return preprocessed

    def keyword_argument(self, tree):
        return self._format_non_option(tree)



import pytest
from gersemi.exceptions import (
    ParsingError,
    UnbalancedParentheses,
    UnbalancedBrackets,
)
from .tests_generator import generate_input_only_tests


def test_parser(parser, case):
    try:
        parser.parse(case.content)
    except ParsingError:
        pytest.fail("invalid input to parse")
        raise


@pytest.mark.parametrize(
    ["invalid_code", "expected_exception"],
    [
        ("set(FOO BAR", UnbalancedParentheses),
        ("message(FOO BAR (BAZ AND FOO)", UnbalancedParentheses),
        ("set(FOO BAR # )", UnbalancedParentheses),
        ("bar)", UnbalancedParentheses),
        ("bar(FOO BAR (BAZ OR FOO)))", UnbalancedParentheses),
        ("another_command #(", UnbalancedParentheses),
        ("foo_command", UnbalancedParentheses),
        ("foo([[foo]=])", UnbalancedBrackets),
        ("foo([=[bar]])", UnbalancedBrackets),
        ("foo(arg1 arg2 [==[arg3]===] arg4)", UnbalancedBrackets),
    ],
)
def test_invalid_code_parsing_error(parser, invalid_code, expected_exception):
    try:
        parser.parse(invalid_code)
        raise AssertionError("Parser should throw an exception")
    except ParsingError as e:
        assert isinstance(e, expected_exception)


pytest_generate_tests = generate_input_only_tests(
    where="parser", input_extension=".cmake"
)



from gersemi.ast_helpers import is_comment
from gersemi.base_dumper import BaseDumper


class KeywordWithPairsFormatter(BaseDumper):
    def _format_pair(self, key, value):
        result = self._try_to_format_into_single_line([key, value])
        if result is not None:
            return result

        formatted_key = self.visit(key)
        with self.indented():
            formatted_value = self.visit(value)
        return "\n".join([formatted_key, formatted_value])

    def _format_keyword_with_pairs(self, args):
        lines = []
        iterator = iter(args)
        for arg in iterator:
            if is_comment(arg):
                lines.append(self.visit(arg))
                continue

            key = arg
            value = next(iterator, None)
            if value is None:
                line = self.visit(key)
            else:
                line = self._format_pair(key, value)
            lines.append(line)
        return "\n".join(lines)



from contextlib import contextmanager
from typing import Iterable, Mapping
from lark import Tree
from gersemi.ast_helpers import (
    get_value,
    is_one_value_argument,
    is_multi_value_argument,
    is_one_of_keywords,
    is_positional_arguments,
    is_section,
    positional_arguments,
)
from gersemi.keyword_kind import kind_to_preprocessor
from gersemi.keywords import KeywordMatcher
from .argument_aware_command_invocation_dumper import (
    ArgumentAwareCommandInvocationDumper,
)


Sections = Mapping[KeywordMatcher, Mapping[KeywordMatcher, Iterable[KeywordMatcher]]]


def create_section_patch(section, old_class):
    def get(key, default_value):
        return section.get(key, default_value)

    class Impl(old_class):
        front_positional_arguments = get("front_positional_arguments", [])
        back_positional_arguments = get("back_positional_arguments", [])
        options = get("options", [])
        one_value_keywords = get("one_value_keywords", [])
        multi_value_keywords = get("multi_value_keywords", [])
        sections = get("sections", dict())
        keyword_formatters = get("keyword_formatters", dict())
        keyword_preprocessors = get("keyword_preprocessors", dict())

    return Impl


class SectionAwareCommandInvocationDumper(ArgumentAwareCommandInvocationDumper):
    sections: Sections = {}

    @contextmanager
    def _update_section_characteristics(self, keyword):
        old_class = type(self)
        try:
            self.__class__ = create_section_patch(self.sections[keyword], old_class)
            yield
        finally:
            self.__class__ = old_class

    def _get_matcher(self, keyword):
        for item in self.sections:
            matcher = is_one_of_keywords([item])
            if matcher(keyword):
                return item
        return None

    def _split_multi_value_argument(self, tree):
        front_node, *rest = tree.children

        keyword_matcher = self._get_matcher(front_node)
        if keyword_matcher is None:
            return tree

        with self._update_section_characteristics(keyword_matcher):
            subarguments = [front_node]
            for arg in self._split_arguments(rest):
                if not is_positional_arguments(arg):
                    subarguments.append(arg)
                    continue

                subarguments.extend(arg.children)

            return Tree("section", subarguments)

    def _is_among_section_keywords(self, section_matcher, argument):
        if section_matcher is None:
            return False

        with self._update_section_characteristics(section_matcher):
            for keywords in (
                self.options,
                self.one_value_keywords,
                self.multi_value_keywords,
            ):
                if is_one_of_keywords(keywords)(argument.children[0]):
                    return True

        return False

    def _fix_back_positional_arguments(self, section):
        if not is_section(section):
            return section

        pivot = len(self.back_positional_arguments)
        if pivot == 0:
            return section

        *front, last = section.children
        if is_one_value_argument(last) or is_multi_value_argument(last):
            last_front, *last_rest = last.children
            left_in_place, back_positional_arguments = (
                last_rest[:-pivot],
                last_rest[-pivot:],
            )
            last.children = [last_front, *left_in_place]
            section.children = [
                *front,
                last,
                positional_arguments(back_positional_arguments),
            ]

        return section

    def _form_sections(self, arguments):
        result = []
        section_matcher = None
        for argument in arguments:
            if self._is_among_section_keywords(section_matcher, argument):
                result[-1].children.append(argument)
            else:
                if section_matcher is not None:
                    with self._update_section_characteristics(section_matcher):
                        result[-1] = self._fix_back_positional_arguments(result[-1])

                if len(argument.children) > 0:
                    section_matcher = self._get_matcher(argument.children[0])
                else:
                    section_matcher = None

                result.append(argument)

        return self._fix_back_positional_arguments(result)

    def _split_arguments(self, arguments):
        preprocessed = super()._split_arguments(arguments)
        return self._form_sections(
            (
                self._split_multi_value_argument(child)
                if is_multi_value_argument(child)
                else child
            )
            for child in preprocessed
        )

    def section(self, tree):
        header, *rest = tree.children
        preprocessor = kind_to_preprocessor(
            self.keyword_preprocessors.get(get_value(header, None), None)
        )
        if preprocessor is not None:

            rest = getattr(self, preprocessor)(rest)

        result = self._try_to_format_into_single_line(tree.children)
        if result is not None:
            return result

        begin = self.visit(header)
        if len(rest) == 0:
            return begin

        with self.indented():
            formatted_values = "\n".join(map(self.visit, rest))
        return f"{begin}\n{formatted_values}"



from gersemi.ast_helpers import is_comment
from gersemi.base_dumper import BaseDumper
from gersemi.utils import pop_all


class KeywordPreprocessor(BaseDumper):
    def _get_bucket_value(self, bucket):
        *comments, node = bucket
        return self.visit(node), tuple(self.visit(comment) for comment in comments)

    def _bucket_arguments_with_their_preceding_comments(self, args):
        result = []
        accumulator = []
        for arg in args:
            accumulator.append(arg)
            if not is_comment(arg):
                result.append(pop_all(accumulator))

        if accumulator:
            result.append(pop_all(accumulator))

        return result

    def _sort_arguments(self, args):
        buckets = self._bucket_arguments_with_their_preceding_comments(args)
        sorted_buckets = sorted(buckets, key=self._get_bucket_value)

        return [arg for bucket in sorted_buckets for arg in bucket]

    def _keep_unique_arguments(self, args):
        buckets = self._bucket_arguments_with_their_preceding_comments(args)
        known = set()
        unique_buckets = []
        for bucket in buckets:
            value = self._get_bucket_value(bucket)
            if value not in known:
                known.add(value)
                unique_buckets.append(bucket)

        return [arg for bucket in unique_buckets for arg in bucket]

    def _sort_and_keep_unique_arguments(self, args):
        return self._sort_arguments(self._keep_unique_arguments(args))



import os
from lark import Lark, UnexpectedInput
from gersemi.exceptions import (
    GenericParsingError,
    UnbalancedParentheses,
    UnbalancedBrackets,
)
from gersemi.parsing_transformer import ParsingTransformer
from gersemi.postprocessor import postprocess


class Parser:
    examples = {
        UnbalancedBrackets: [
            "foo(foo [[foo]=]",
            "foo([=[foo bar]])",
            "foo([=[foo bar ]==])",
            "foo(foo [=[foo bar ]==] foo)",
            "foo(foo foo [==[foo]===] foo)",
        ],
        UnbalancedParentheses: [
            "foo(foo foo",
            "foo(foo foo # )",
            "foo(bar",
            "foo(bar\n",
            "foo(BAR (BAZ)",
            "foo(# )",
            "foo(bar))",
            "foo)",
            "foo(BAR (BAZ)))",
            "foo(BAR (BAZ FOO)))",
            "foo",
            "foo # (",
            "foo(commented_argument #foobar)",
            "foo(commented_argument #[[foobar]]",
            "foo(almost_commented_argument #)",
        ],
    }

    def __init__(self, grammar_filename):
        self.lark_parser = Lark.open(
            grammar_filename=grammar_filename,
            parser="lalr",
            propagate_positions=False,
            maybe_placeholders=False,
            transformer=ParsingTransformer(),
        )

    def _match_parsing_error(self, code, exception):
        specific_error = exception.match_examples(self.lark_parser.parse, self.examples)
        if not specific_error:
            raise GenericParsingError(
                exception.get_context(code), exception.line, exception.column
            )
        raise specific_error(
            exception.get_context(code), exception.line, exception.column
        )

    def parse(self, code):  # pylint: disable=inconsistent-return-statements
        try:
            return self.lark_parser.parse(code)
        except UnexpectedInput as u:
            self._match_parsing_error(code, u)


class ParserWithPostProcessing:
    def __init__(self, parser):
        self.parser = parser

    def parse(self, code, known_definitions=None):
        return postprocess(code, known_definitions, self.parser.parse(code))


HERE = os.path.dirname(os.path.realpath(__file__))
GRAMMAR = os.path.join(HERE, "cmake.lark")


def create_parser(grammar_filename=GRAMMAR):
    return Parser(grammar_filename)


def create_parser_with_postprocessing(bare_parser):
    return ParserWithPostProcessing(bare_parser)


BARE_PARSER = create_parser()
PARSER = create_parser_with_postprocessing(BARE_PARSER)



from pathlib import Path
from typing import Union


class ModuleExtension(str):
    @property
    def name(self):
        return str(self)

    @property
    def qualified_name(self):
        return f"gersemi_{self}"


class FileExtension(str):
    def __init__(self, value):
        super().__init__()
        self.path = Path(value).resolve(True)

    @property
    def name(self):
        return str(self.path)

    @property
    def qualified_name(self):
        return self.name


Extension = Union[ModuleExtension, FileExtension]
